@@ -3,12 +3,13 @@ from __future__ import unicode_literals
 from prompt_toolkit.buffer import SelectionType, indent, unindent
 from prompt_toolkit.keys import Keys
 from prompt_toolkit.enums import IncrementalSearchDirection, SEARCH_BUFFER, SYSTEM_BUFFER
-from prompt_toolkit.filters import CLIFilter, Always
+from prompt_toolkit.filters import Always, Condition, EmacsMode, to_cli_filter, HasSelection, EmacsInsertMode, HasFocus
+from prompt_toolkit.completion import CompleteEvent
 
 from .utils import create_handle_decorator
 from .scroll import scroll_page_up, scroll_page_down
 
-import prompt_toolkit.filters as filters
+from six.moves import range
 
 __all__ = (
     'load_emacs_bindings',
@@ -24,11 +25,11 @@ def load_emacs_bindings(registry, filter=Always()):
     """
     # Overview of Readline emacs commands:
     # http://www.catonmat.net/download/readline-emacs-editing-mode-cheat-sheet.pdf
+    filter = to_cli_filter(filter)
 
-    assert isinstance(filter, CLIFilter)
-
-    handle = create_handle_decorator(registry, filter)
-    has_selection = filters.HasSelection()
+    handle = create_handle_decorator(registry, filter & EmacsMode())
+    insert_mode = EmacsInsertMode()
+    has_selection = HasSelection()
 
     @handle(Keys.Escape)
     def _(event):
@@ -88,7 +89,7 @@ def load_emacs_bindings(registry, filter=Always()):
         """
         event.current_buffer.cursor_down()
 
-    @handle(Keys.ControlO, filter= ~has_selection)
+    @handle(Keys.ControlO, filter=insert_mode)
     def _(event):
         """
         Insert newline, but don't move the cursor.
@@ -120,8 +121,8 @@ def load_emacs_bindings(registry, filter=Always()):
         """
         event.current_buffer.insert_text(event.data, overwrite=False)
 
-    @handle(Keys.ControlY, filter= ~has_selection)
-    @handle(Keys.ControlX, 'r', 'y', filter= ~has_selection)
+    @handle(Keys.ControlY, filter=insert_mode)
+    @handle(Keys.ControlX, 'r', 'y', filter=insert_mode)
     def _(event):
         """
         Paste before cursor.
@@ -129,7 +130,7 @@ def load_emacs_bindings(registry, filter=Always()):
         event.current_buffer.paste_clipboard_data(
             event.cli.clipboard.get_data(), count=event.arg, before=True)
 
-    @handle(Keys.ControlUnderscore, save_before=(lambda e: False), filter= ~has_selection)
+    @handle(Keys.ControlUnderscore, save_before=(lambda e: False), filter=insert_mode)
     def _(event):
         """
         Undo.
@@ -154,15 +155,16 @@ def load_emacs_bindings(registry, filter=Always()):
         if event._arg is None:
             event.append_to_arg_count('-')
 
-    @handle(Keys.Escape, Keys.ControlJ, filter= ~has_selection)
+    is_returnable = Condition(
+        lambda cli: cli.current_buffer.accept_action.is_returnable)
+
+    @handle(Keys.Escape, Keys.ControlJ, filter=insert_mode & is_returnable)
     def _(event):
         """
         Meta + Newline: always accept input.
         """
         b = event.current_buffer
-
-        if b.accept_action.is_returnable:
-            b.accept_action.validate_and_handle(event.cli, b)
+        b.accept_action.validate_and_handle(event.cli, b)
 
     @handle(Keys.ControlSquareClose, Keys.Any)
     def _(event):
@@ -173,7 +175,7 @@ def load_emacs_bindings(registry, filter=Always()):
         if match is not None:
             event.current_buffer.cursor_position += match
 
-    @handle(Keys.Escape, Keys.Backspace, filter= ~has_selection)
+    @handle(Keys.Escape, Keys.Backspace, filter=insert_mode)
     def _(event):
         """
         Delete word backwards.
@@ -181,11 +183,27 @@ def load_emacs_bindings(registry, filter=Always()):
         buffer = event.current_buffer
         pos = buffer.document.find_start_of_previous_word(count=event.arg)
 
+        if pos is None:
+            # Nothing found. Only whitespace before the cursor?
+            pos = - buffer.cursor_position
+
         if pos:
             deleted = buffer.delete_before_cursor(count=-pos)
             event.cli.clipboard.set_text(deleted)
 
-    @handle(Keys.Escape, 'a', filter= ~has_selection)
+    @handle(Keys.ControlDelete, filter=insert_mode)
+    def _(event):
+        """
+        Delete word after cursor.
+        """
+        buff = event.current_buffer
+        pos = buff.document.find_next_word_ending(count=event.arg)
+
+        if pos:
+            deleted = buff.delete(count=pos)
+            event.cli.clipboard.set_text(deleted)
+
+    @handle(Keys.Escape, 'a')
     def _(event):
         """
         Previous sentence.
@@ -193,7 +211,7 @@ def load_emacs_bindings(registry, filter=Always()):
         # TODO:
         pass
 
-    @handle(Keys.Escape, 'c', filter= ~has_selection)
+    @handle(Keys.Escape, 'c', filter=insert_mode)
     def _(event):
         """
         Capitalize the current (or following) word.
@@ -205,7 +223,7 @@ def load_emacs_bindings(registry, filter=Always()):
             words = buffer.document.text_after_cursor[:pos]
             buffer.insert_text(words.title(), overwrite=True)
 
-    @handle(Keys.Escape, 'd', filter= ~has_selection)
+    @handle(Keys.Escape, 'd', filter=insert_mode)
     def _(event):
         """
         Delete word forwards.
@@ -217,7 +235,7 @@ def load_emacs_bindings(registry, filter=Always()):
             deleted = buffer.delete(count=pos)
             event.cli.clipboard.set_text(deleted)
 
-    @handle(Keys.Escape, 'e', filter= ~has_selection)
+    @handle(Keys.Escape, 'e')
     def _(event):
         """ Move to end of sentence. """
         # TODO:
@@ -246,7 +264,7 @@ def load_emacs_bindings(registry, filter=Always()):
         if pos:
             buffer.cursor_position += pos
 
-    @handle(Keys.Escape, 'l', filter= ~has_selection)
+    @handle(Keys.Escape, 'l', filter=insert_mode)
     def _(event):
         """
         Lowercase the current (or following) word.
@@ -258,14 +276,14 @@ def load_emacs_bindings(registry, filter=Always()):
             words = buffer.document.text_after_cursor[:pos]
             buffer.insert_text(words.lower(), overwrite=True)
 
-    @handle(Keys.Escape, 't', filter= ~has_selection)
+    @handle(Keys.Escape, 't', filter=insert_mode)
     def _(event):
         """
         Swap the last two words before the cursor.
         """
         # TODO
 
-    @handle(Keys.Escape, 'u', filter= ~has_selection)
+    @handle(Keys.Escape, 'u', filter=insert_mode)
     def _(event):
         """
         Uppercase the current (or following) word.
@@ -277,27 +295,45 @@ def load_emacs_bindings(registry, filter=Always()):
             words = buffer.document.text_after_cursor[:pos]
             buffer.insert_text(words.upper(), overwrite=True)
 
-    @handle(Keys.Escape, '.', filter= ~has_selection)
+    @handle(Keys.Escape, '.', filter=insert_mode)
     def _(event):
         """
         Rotate through the last word (white-space delimited) of the previous lines in history.
         """
         # TODO
 
-    @handle(Keys.Escape, '\\', filter= ~has_selection)
+    @handle(Keys.Escape, '\\', filter=insert_mode)
     def _(event):
         """
         Delete all spaces and tabs around point.
         (delete-horizontal-space)
         """
+        buff = event.current_buffer
+        text_before_cursor = buff.document.text_before_cursor
+        text_after_cursor = buff.document.text_after_cursor
 
-    @handle(Keys.Escape, '*', filter= ~has_selection)
+        delete_before = len(text_before_cursor) - len(text_before_cursor.rstrip('\t '))
+        delete_after = len(text_after_cursor) - len(text_after_cursor.lstrip('\t '))
+
+        buff.delete_before_cursor(count=delete_before)
+        buff.delete(count=delete_after)
+
+    @handle(Keys.Escape, '*', filter=insert_mode)
     def _(event):
         """
         `meta-*`: Insert all possible completions of the preceding text.
         """
+        buff = event.current_buffer
 
-    @handle(Keys.ControlX, Keys.ControlU, save_before=(lambda e: False), filter= ~has_selection)
+        # List all completions.
+        complete_event = CompleteEvent(text_inserted=False, completion_requested=True)
+        completions = list(buff.completer.get_completions(buff.document, complete_event))
+
+        # Insert them.
+        text_to_insert = ' '.join(c.text for c in completions)
+        buff.insert_text(text_to_insert)
+
+    @handle(Keys.ControlX, Keys.ControlU, save_before=(lambda e: False), filter=insert_mode)
     def _(event):
         event.current_buffer.undo()
 
@@ -317,10 +353,12 @@ def load_emacs_bindings(registry, filter=Always()):
     @handle(Keys.ControlSpace)
     def _(event):
         """
-        Start of the selection.
+        Start of the selection (if the current buffer is not empty).
         """
         # Take the current cursor position as the start of this selection.
-        event.current_buffer.start_selection(selection_type=SelectionType.CHARACTERS)
+        buff = event.current_buffer
+        if buff.text:
+            buff.start_selection(selection_type=SelectionType.CHARACTERS)
 
     @handle(Keys.ControlG, filter= ~has_selection)
     def _(event):
@@ -387,7 +425,7 @@ def load_emacs_bindings(registry, filter=Always()):
         buffer.cursor_position += buffer.document.find_next_word_beginning(count=event.arg) or \
             buffer.document.get_end_of_document_position()
 
-    @handle(Keys.Escape, '/', filter= ~has_selection)
+    @handle(Keys.Escape, '/', filter=insert_mode)
     def _(event):
         """
         M-/: Complete.
@@ -411,7 +449,7 @@ def load_emacs_bindings(registry, filter=Always()):
         from_, _ = buffer.document.translate_index_to_position(from_)
         to, _ = buffer.document.translate_index_to_position(to)
 
-        indent(buffer, from_ - 1, to, count=event.arg)  # XXX: why does translate_index_to_position return 1-based indexing???
+        indent(buffer, from_, to + 1, count=event.arg)
 
     @handle(Keys.ControlC, '<', filter=has_selection)
     def _(event):
@@ -424,7 +462,7 @@ def load_emacs_bindings(registry, filter=Always()):
         from_, _ = buffer.document.translate_index_to_position(from_)
         to, _ = buffer.document.translate_index_to_position(to)
 
-        unindent(buffer, from_ - 1, to, count=event.arg)
+        unindent(buffer, from_, to + 1, count=event.arg)
 
 
 def load_emacs_open_in_editor_bindings(registry, filter=None):
@@ -432,7 +470,7 @@ def load_emacs_open_in_editor_bindings(registry, filter=None):
     Pressing C-X C-E will open the buffer in an external editor.
     """
     handle = create_handle_decorator(registry, filter)
-    has_selection = filters.HasSelection()
+    has_selection = HasSelection()
 
     @handle(Keys.ControlX, Keys.ControlE, filter= ~has_selection)
     def _(event):
@@ -444,14 +482,14 @@ def load_emacs_open_in_editor_bindings(registry, filter=None):
 
 def load_emacs_system_bindings(registry, filter=None):
     handle = create_handle_decorator(registry, filter)
-    has_focus = filters.HasFocus(SYSTEM_BUFFER)
+    has_focus = HasFocus(SYSTEM_BUFFER)
 
     @handle(Keys.Escape, '!', filter= ~has_focus)
     def _(event):
         """
         M-'!' opens the system prompt.
         """
-        event.cli.focus_stack.push(SYSTEM_BUFFER)
+        event.cli.push_focus(SYSTEM_BUFFER)
 
     @handle(Keys.Escape, filter=has_focus)
     @handle(Keys.ControlG, filter=has_focus)
@@ -461,7 +499,7 @@ def load_emacs_system_bindings(registry, filter=None):
         Cancel system prompt.
         """
         event.cli.buffers[SYSTEM_BUFFER].reset()
-        event.cli.focus_stack.pop()
+        event.cli.pop_focus()
 
     @handle(Keys.ControlJ, filter=has_focus)
     def _(event):
@@ -473,12 +511,19 @@ def load_emacs_system_bindings(registry, filter=None):
         system_line.reset(append_to_history=True)
 
         # Focus previous buffer again.
-        event.cli.focus_stack.pop()
+        event.cli.pop_focus()
 
 
-def load_emacs_search_bindings(registry, filter=None):
-    handle = create_handle_decorator(registry, filter)
-    has_focus = filters.HasFocus(SEARCH_BUFFER)
+def load_emacs_search_bindings(registry, get_search_state=None, filter=None):
+    filter = to_cli_filter(filter)
+
+    handle = create_handle_decorator(registry, filter & EmacsMode())
+    has_focus = HasFocus(SEARCH_BUFFER)
+
+    assert get_search_state is None or callable(get_search_state)
+
+    if not get_search_state:
+        def get_search_state(cli): return cli.search_state
 
     @handle(Keys.ControlG, filter=has_focus)
     @handle(Keys.ControlC, filter=has_focus)
@@ -491,7 +536,7 @@ def load_emacs_search_bindings(registry, filter=None):
         search_buffer = event.cli.buffers[SEARCH_BUFFER]
 
         search_buffer.reset()
-        event.cli.focus_stack.pop()
+        event.cli.pop_focus()
 
     @handle(Keys.ControlJ, filter=has_focus)
     def _(event):
@@ -499,64 +544,57 @@ def load_emacs_search_bindings(registry, filter=None):
         When enter pressed in isearch, quit isearch mode. (Multiline
         isearch would be too complicated.)
         """
-        input_buffer = event.cli.buffers[event.cli.focus_stack.previous]
+        input_buffer = event.cli.buffers.previous(event.cli)
         search_buffer = event.cli.buffers[SEARCH_BUFFER]
 
         # Update search state.
         if search_buffer.text:
-            event.cli.search_state.text = search_buffer.text
+            get_search_state(event.cli).text = search_buffer.text
 
         # Apply search.
-        input_buffer.apply_search(event.cli.search_state, include_current_position=True)
+        input_buffer.apply_search(get_search_state(event.cli), include_current_position=True)
 
         # Add query to history of search line.
         search_buffer.append_to_history()
         search_buffer.reset()
 
         # Focus previous document again.
-        event.cli.focus_stack.pop()
+        event.cli.pop_focus()
 
     @handle(Keys.ControlR, filter= ~has_focus)
     def _(event):
-        event.cli.search_state.direction = IncrementalSearchDirection.BACKWARD
-        event.cli.focus_stack.push(SEARCH_BUFFER)
+        get_search_state(event.cli).direction = IncrementalSearchDirection.BACKWARD
+        event.cli.push_focus(SEARCH_BUFFER)
 
     @handle(Keys.ControlS, filter= ~has_focus)
     def _(event):
-        event.cli.search_state.direction = IncrementalSearchDirection.FORWARD
-        event.cli.focus_stack.push(SEARCH_BUFFER)
+        get_search_state(event.cli).direction = IncrementalSearchDirection.FORWARD
+        event.cli.push_focus(SEARCH_BUFFER)
+
+    def incremental_search(cli, direction, count=1):
+        " Apply search, but keep search buffer focussed. "
+        # Update search_state.
+        search_state = get_search_state(cli)
+        direction_changed = search_state.direction != direction
+
+        search_state.text = cli.buffers[SEARCH_BUFFER].text
+        search_state.direction = direction
+
+        # Apply search to current buffer.
+        if not direction_changed:
+            input_buffer = cli.buffers.previous(cli)
+            input_buffer.apply_search(search_state,
+                                      include_current_position=False, count=count)
 
     @handle(Keys.ControlR, filter=has_focus)
     @handle(Keys.Up, filter=has_focus)
     def _(event):
-        # Update search_state.
-        search_state = event.cli.search_state
-        direction_changed = search_state.direction != IncrementalSearchDirection.BACKWARD
-
-        search_state.text = event.cli.buffers[SEARCH_BUFFER].text
-        search_state.direction = IncrementalSearchDirection.BACKWARD
-
-        # Apply search to current buffer.
-        if not direction_changed:
-            input_buffer = event.cli.buffers[event.cli.focus_stack.previous]
-            input_buffer.apply_search(event.cli.search_state,
-                                      include_current_position=False, count=event.arg)
+        incremental_search(event.cli, IncrementalSearchDirection.BACKWARD, count=event.arg)
 
     @handle(Keys.ControlS, filter=has_focus)
     @handle(Keys.Down, filter=has_focus)
     def _(event):
-        # Update search_state.
-        search_state = event.cli.search_state
-        direction_changed = search_state.direction != IncrementalSearchDirection.FORWARD
-
-        search_state.text = event.cli.buffers[SEARCH_BUFFER].text
-        search_state.direction = IncrementalSearchDirection.FORWARD
-
-        # Apply search to current buffer.
-        if not direction_changed:
-            input_buffer = event.cli.buffers[event.cli.focus_stack.previous]
-            input_buffer.apply_search(event.cli.search_state,
-                                      include_current_position=False, count=event.arg)
+        incremental_search(event.cli, IncrementalSearchDirection.FORWARD, count=event.arg)
 
 
 def load_extra_emacs_page_navigation_bindings(registry, filter=None):
@@ -564,7 +602,8 @@ def load_extra_emacs_page_navigation_bindings(registry, filter=None):
     Key bindings, for scrolling up and down through pages.
     This are separate bindings, because GNU readline doesn't have them.
     """
-    handle = create_handle_decorator(registry, filter)
+    filter = to_cli_filter(filter)
+    handle = create_handle_decorator(registry, filter & EmacsMode())
 
     handle(Keys.ControlV)(scroll_page_down)
     handle(Keys.PageDown)(scroll_page_down)
